@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
+ * vim: set ts=2 sw=2 et tw=78:
  *
  * ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
@@ -70,7 +71,7 @@
 #include "nsNetUtil.h"
 #include "nsIScriptSecurityManager.h"
 #include "nsDOMError.h"
-#include "nsIDOMWindowInternal.h"
+#include "nsPIDOMWindow.h"
 #include "nsIJSContextStack.h"
 #include "nsIDocShell.h"
 #include "nsIDocShellTreeItem.h"
@@ -86,6 +87,7 @@
 #include "nsIForm.h"
 #include "nsIFormControl.h"
 #include "nsHTMLAtoms.h"
+#include "nsHTMLParts.h"
 #include "nsISupportsPrimitives.h"
 #include "nsLayoutAtoms.h"
 #include "imgIDecoderObserver.h"
@@ -111,6 +113,8 @@ static NS_DEFINE_CID(kXTFServiceCID, NS_XTFSERVICE_CID);
 #include "nsIMIMEService.h"
 #include "jsdbgapi.h"
 #include "nsIJSRuntimeService.h"
+#include "nsIFragmentContentSink.h"
+#include "nsIScriptObjectPrincipal.h"
 
 // for ReportToConsole
 #include "nsIStringBundle.h"
@@ -415,6 +419,9 @@ nsContentUtils::Shutdown()
 {
   sInitialized = PR_FALSE;
 
+  NS_HTMLParanoidFragmentSinkShutdown();
+  NS_XHTMLParanoidFragmentSinkShutdown();
+
   PRInt32 i;
   for (i = 0; i < PRInt32(PropertiesFile_COUNT); ++i)
     NS_IF_RELEASE(sStringBundles[i]);
@@ -638,6 +645,17 @@ nsContentUtils::CheckSameOrigin(nsIDOMNode *aTrustedNode,
     }
   }
 
+  if (!unTrustedPrincipal) {
+    unTrustedPrincipal = unTrustedDoc->GetPrincipal();
+
+    if (!unTrustedPrincipal) {
+      // If the trusted node doesn't have a principal we can't check
+      // security against it
+
+      return NS_ERROR_DOM_SECURITY_ERR;
+    }
+  }
+
   return sSecurityManager->CheckSameOriginPrincipal(trustedPrincipal,
                                                     unTrustedPrincipal);
 }
@@ -682,8 +700,17 @@ nsContentUtils::CanCallerAccess(nsIDOMNode *aNode)
     return PR_TRUE;
   }
 
-  rv = sSecurityManager->CheckSameOriginPrincipal(subjectPrincipal,
-                                                  principal);
+  PRBool enabled = PR_FALSE;
+  if (principal == systemPrincipal) {
+    // we know subjectPrincipal != systemPrincipal so we can only
+    // access the object if UniversalXPConnect is enabled. We can
+    // avoid wasting time in CheckSameOriginPrincipal
+
+    rv = sSecurityManager->IsCapabilityEnabled("UniversalXPConnect", &enabled);
+    return NS_SUCCEEDED(rv) && enabled;
+  }
+
+  rv = sSecurityManager->CheckSameOriginPrincipal(subjectPrincipal, principal);
   if (NS_SUCCEEDED(rv)) {
     return PR_TRUE;
   }
@@ -691,11 +718,60 @@ nsContentUtils::CanCallerAccess(nsIDOMNode *aNode)
   // see if the caller has otherwise been given the ability to touch
   // input args to DOM methods
 
+  rv = sSecurityManager->IsCapabilityEnabled("UniversalBrowserRead", &enabled);
+  return NS_SUCCEEDED(rv) && enabled;
+}
+
+// static
+PRBool
+nsContentUtils::CanCallerAccess(nsPIDOMWindow* aWindow)
+{
+  nsCOMPtr<nsIPrincipal> subjectPrincipal;
+  sSecurityManager->GetSubjectPrincipal(getter_AddRefs(subjectPrincipal));
+
+  if (!subjectPrincipal) {
+    // we're running as system, grant access to the node.
+
+    return PR_TRUE;
+  }
+
+  nsCOMPtr<nsIPrincipal> systemPrincipal;
+  sSecurityManager->GetSystemPrincipal(getter_AddRefs(systemPrincipal));
+
+  if (subjectPrincipal == systemPrincipal) {
+    // we're running as system, grant access to the node.
+
+    return PR_TRUE;
+  }
+
+  nsCOMPtr<nsIScriptObjectPrincipal> scriptObject =
+    do_QueryInterface(aWindow->IsOuterWindow() ?
+                      aWindow->GetCurrentInnerWindow() : aWindow);
+  NS_ENSURE_TRUE(scriptObject, PR_FALSE);
+  nsIPrincipal* principal = scriptObject->GetPrincipal();
+  NS_ENSURE_TRUE(principal, PR_FALSE);
+
   PRBool enabled = PR_FALSE;
-  rv = sSecurityManager->IsCapabilityEnabled("UniversalBrowserRead",
-                                             &enabled);
-  NS_ENSURE_SUCCESS(rv, PR_FALSE);
-  return enabled;
+  nsresult rv;
+  if (principal == systemPrincipal) {
+    // we know subjectPrincipal != systemPrincipal so we can only
+    // access the object if UniversalXPConnect is enabled. We can
+    // avoid wasting time in CheckSameOriginPrincipal
+
+    rv = sSecurityManager->IsCapabilityEnabled("UniversalXPConnect", &enabled);
+    return NS_SUCCEEDED(rv) && enabled;
+  }
+
+  rv = sSecurityManager->CheckSameOriginPrincipal(subjectPrincipal, principal);
+  if (NS_SUCCEEDED(rv)) {
+    return PR_TRUE;
+  }
+
+  // see if the caller has otherwise been given the ability to touch
+  // input args to DOM methods
+
+  rv = sSecurityManager->IsCapabilityEnabled("UniversalBrowserRead", &enabled);
+  return NS_SUCCEEDED(rv) && enabled;
 }
 
 //static
@@ -739,38 +815,33 @@ nsContentUtils::InProlog(nsIDOMNode *aNode)
 
 // static
 nsresult
-nsContentUtils::doReparentContentWrapper(nsIContent *aChild,
+nsContentUtils::doReparentContentWrapper(nsIContent *aNode,
                                          JSContext *cx,
                                          JSObject *aOldGlobal,
-                                         JSObject *parent_obj)
+                                         JSObject *aNewGlobal)
 {
   nsCOMPtr<nsIXPConnectJSObjectHolder> old_wrapper;
 
   nsresult rv;
 
-  rv = sXPConnect->ReparentWrappedNativeIfFound(cx, aOldGlobal, parent_obj,
-                                                aChild,
+  rv = sXPConnect->ReparentWrappedNativeIfFound(cx, aOldGlobal, aNewGlobal,
+                                                aNode,
                                                 getter_AddRefs(old_wrapper));
   NS_ENSURE_SUCCESS(rv, rv);
 
-  if (!old_wrapper) {
-    // If aChild isn't wrapped none of it's children are wrapped so
-    // there's no need to walk into aChild's children.
+  // Whether or not aChild is already wrapped we must iterate through
+  // its descendants since there's no guarantee that a descendant isn't
+  // wrapped even if this child is not wrapped. That used to be true
+  // when every DOM node's JSObject was parented at the DOM node's
+  // parent's JSObject, but that's no longer the case.
 
-    return NS_OK;
-  }
-
-  JSObject *old;
-  rv = old_wrapper->GetJSObject(&old);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  PRUint32 i, count = aChild->GetChildCount();
+  PRUint32 i, count = aNode->GetChildCount();
 
   for (i = 0; i < count; i++) {
-    nsIContent *child = aChild->GetChildAt(i);
+    nsIContent *child = aNode->GetChildAt(i);
     NS_ENSURE_TRUE(child, NS_ERROR_UNEXPECTED);
 
-    rv = doReparentContentWrapper(child, cx, aOldGlobal, old);
+    rv = doReparentContentWrapper(child, cx, aOldGlobal, aNewGlobal);
     NS_ENSURE_SUCCESS(rv, rv);
   }
 
@@ -780,18 +851,24 @@ nsContentUtils::doReparentContentWrapper(nsIContent *aChild,
 static JSContext *
 GetContextFromDocument(nsIDocument *aDocument, JSObject** aGlobalObject)
 {
-  nsIScriptGlobalObject *sgo = aDocument->GetScriptGlobalObject();
-
+  nsCOMPtr<nsIDocument_MOZILLA_1_8_0_BRANCH> doc18 =
+    do_QueryInterface(aDocument);
+  if (!doc18) {
+    NS_ERROR("What kind of crazy document was passed in?");
+    return nsnull;
+  }
+  nsIScriptGlobalObject *sgo = doc18->GetScopeObject();
   if (!sgo) {
     // No script global, no context.
 
+    *aGlobalObject = nsnull;
+
     return nsnull;
   }
 
   *aGlobalObject = sgo->GetGlobalJSObject();
-  
-  nsIScriptContext *scx = sgo->GetContext();
 
+  nsIScriptContext *scx = sgo->GetContext();
   if (!scx) {
     // No context left in the old scope...
 
@@ -812,18 +889,25 @@ nsContentUtils::ReparentContentWrapper(nsIContent *aContent,
     return NS_OK;
   }
 
-  if (!aOldDocument) {
-    // If we can't find our old document we don't know what our old
-    // scope was so there's no way to find the old wrapper
 
+  nsCOMPtr<nsIDocument_MOZILLA_1_8_0_BRANCH> doc18 =
+    do_QueryInterface(aNewDocument);
+  if (!doc18) {
+    NS_ERROR("Crazy document passed in");
+    return NS_ERROR_UNEXPECTED;
+  }
+  nsIScriptGlobalObject *newSGO = doc18->GetScopeObject();
+  JSObject *newScope;
+
+  // If we can't find our old document we don't know what our old
+  // scope was so there's no way to find the old wrapper, and if there
+  // is no new scope there's no reason to reparent.
+  if (!aOldDocument || !newSGO || !(newScope = newSGO->GetGlobalJSObject())) {
     return NS_OK;
   }
 
   NS_ENSURE_TRUE(sXPConnect, NS_ERROR_NOT_INITIALIZED);
 
-  nsISupports* new_parent = aNewParent ? (nsISupports*)aNewParent :
-    (nsISupports*)aNewDocument;
-
   // Make sure to get our hands on the right scope object, since
   // GetWrappedNativeOfNativeObject doesn't call PreCreate and hence won't get
   // the right scope if we pass in something bogus.  The right scope lives on
@@ -834,41 +918,86 @@ nsContentUtils::ReparentContentWrapper(nsIContent *aContent,
   JSObject *globalObj;
   JSContext *cx = GetContextFromDocument(aOldDocument, &globalObj);
 
-  if (!cx || !globalObj) {
-    // No JSContext left in the old scope, or no global object around; can't
-    // find the old wrapper w/o the old context or global object
+  if (!globalObj) {
+    // No global object around; can't find the old wrapper w/o the old
+    // global object
 
     return NS_OK;
   }
 
-  nsCOMPtr<nsIXPConnectWrappedNative> wrapper;
-  nsresult rv =
-    sXPConnect->GetWrappedNativeOfNativeObject(cx, globalObj, aContent,
-                                               NS_GET_IID(nsISupports),
-                                               getter_AddRefs(wrapper));
-  NS_ENSURE_SUCCESS(rv, rv);
+  if (!cx) {
+    JSObject *dummy;
+    cx = GetContextFromDocument(aNewDocument, &dummy);
 
-  if (!wrapper) {
-    // aContent is not wrapped (and thus none of its children are
-    // wrapped) so there's no need to reparent anything.
+    if (!cx) {
+      // No context reachable from the old or new document, use the
+      // calling context, or the safe context if no caller can be
+      // found.
 
-    return NS_OK;
+      sThreadJSContextStack->Peek(&cx);
+
+      if (!cx) {
+        sThreadJSContextStack->GetSafeJSContext(&cx);
+      }
+    }
   }
 
-  // Wrap the new parent and reparent aContent.  Don't bother using globalObj
-  // here, since it's wrong for the new parent anyway...  Luckily, WrapNative
-  // will PreCreate and hence get the right scope.
-  nsCOMPtr<nsIXPConnectJSObjectHolder> holder;
-  rv = sXPConnect->WrapNative(cx, ::JS_GetGlobalObject(cx), new_parent,
-                              NS_GET_IID(nsISupports),
-                              getter_AddRefs(holder));
-  NS_ENSURE_SUCCESS(rv, rv);
+  return doReparentContentWrapper(aContent, cx, globalObj, newScope);
+}
 
-  JSObject *obj;
-  rv = holder->GetJSObject(&obj);
-  NS_ENSURE_SUCCESS(rv, rv);
+nsresult
+nsContentUtils::ReparentContentWrappersInScope(nsIScriptGlobalObject *aOldScope,
+                                               nsIScriptGlobalObject *aNewScope)
+{
+  JSContext *cx = nsnull;
 
-  return doReparentContentWrapper(aContent, cx, globalObj, obj);
+  // Try really hard to find a context to work on.
+  nsIScriptContext *context = aOldScope->GetContext();
+  if (context) {
+    cx = NS_STATIC_CAST(JSContext *, context->GetNativeContext());
+  }
+
+  if (!cx) {
+    context = aNewScope->GetContext();
+    if (context) {
+      cx = NS_STATIC_CAST(JSContext *, context->GetNativeContext());
+    }
+
+    if (!cx) {
+      sThreadJSContextStack->Peek(&cx);
+
+      if (!cx) {
+        sThreadJSContextStack->GetSafeJSContext(&cx);
+
+        if (!cx) {
+          // Wow, this is really bad!
+          NS_WARNING("No context reachable in ReparentContentWrappers()!");
+
+          return NS_ERROR_NOT_AVAILABLE;
+        }
+      }
+    }
+  }
+
+  // Now that we have a context, let's get the global objects from the two
+  // scopes and ask XPConnect to do the rest of the work.
+
+  JSObject *oldScopeObj = aOldScope->GetGlobalJSObject();
+  JSObject *newScopeObj = aNewScope->GetGlobalJSObject();
+
+  if (!newScopeObj || !oldScopeObj) {
+    // We can't really do anything without the JSObjects.
+
+    return NS_ERROR_NOT_AVAILABLE;
+  }
+
+  nsCOMPtr<nsIXPConnect_MOZILLA_1_8_BRANCH> xpconnect18 =
+    do_QueryInterface(sXPConnect);
+  if (!xpconnect18) {
+    NS_ERROR("Weird things are happening in XPConnect");
+    return NS_ERROR_FAILURE;
+  }
+  return xpconnect18->ReparentScopeAwareWrappers(cx, oldScopeObj, newScopeObj);
 }
 
 nsIDocShell *
@@ -894,6 +1023,31 @@ nsContentUtils::GetDocShellFromCaller()
 
 nsIDOMDocument *
 nsContentUtils::GetDocumentFromCaller()
+{
+  JSContext *cx = nsnull;
+  sThreadJSContextStack->Peek(&cx);
+
+  nsIDOMDocument *doc = nsnull;
+
+  if (cx) {
+    JSObject *callee = nsnull;
+    JSStackFrame *fp = nsnull;
+    while (!callee && (fp = ::JS_FrameIterator(cx, &fp))) {
+      callee = ::JS_GetFrameCalleeObject(cx, fp);
+    }
+
+    nsCOMPtr<nsPIDOMWindow> win =
+      do_QueryInterface(nsJSUtils::GetStaticScriptGlobal(cx, callee));
+    if (win) {
+      doc = win->GetExtantDocument();
+    }
+  }
+
+  return doc;
+}
+
+nsIDOMDocument *
+nsContentUtils::GetDocumentFromContext()
 {
   if (!sThreadJSContextStack) {
     return nsnull;
@@ -922,15 +1076,19 @@ PRBool
 nsContentUtils::IsCallerChrome()
 {
   PRBool is_caller_chrome = PR_FALSE;
-  nsresult rv = sSecurityManager->SubjectPrincipalIsSystem(&is_caller_chrome);
-  if (NS_FAILED(rv)) {
-    return PR_FALSE;
+
+  if (sSecurityManager) {
+    nsresult rv = sSecurityManager->SubjectPrincipalIsSystem(&is_caller_chrome);
+    if (NS_FAILED(rv)) {
+      return PR_FALSE;
+    }
   }
 
   return is_caller_chrome;
 }
 
-static PRBool IsCallerTrustedForCapability(const char* aCapability)
+PRBool
+nsContentUtils::IsCallerTrustedForCapability(const char* aCapability)
 {
   if (nsContentUtils::IsCallerChrome())
     return PR_TRUE;
@@ -2158,13 +2316,13 @@ nsContentUtils::GetEventArgName(PRInt32 aNameSpaceID)
   return gEventName;
 }
 
-void
+PRBool
 nsCxPusher::Push(nsISupports *aCurrentTarget)
 {
   if (mScx) {
     NS_ERROR("Whaaa! No double pushing with nsCxPusher::Push()!");
 
-    return;
+    return PR_FALSE;
   }
 
   nsCOMPtr<nsIScriptGlobalObject> sgo;
@@ -2172,7 +2330,8 @@ nsCxPusher::Push(nsISupports *aCurrentTarget)
   nsCOMPtr<nsIDocument> document;
 
   if (content) {
-    document = content->GetDocument();
+    document = content->GetOwnerDoc();
+    NS_ENSURE_TRUE(document, PR_FALSE);
   }
 
   if (!document) {
@@ -2180,7 +2339,15 @@ nsCxPusher::Push(nsISupports *aCurrentTarget)
   }
 
   if (document) {
-    sgo = document->GetScriptGlobalObject();
+    nsCOMPtr<nsIDocument_MOZILLA_1_8_BRANCH3> branch3doc =
+      do_QueryInterface(document);
+    NS_ASSERTION(branch3doc,
+                 "Document must implement nsIDocument_MOZILLA_1_8_BRANCH3!!!");
+    PRBool hasHadScriptObject = PR_TRUE;
+    sgo = branch3doc->GetScriptHandlingObject(hasHadScriptObject);
+    // It is bad if the document doesn't have event handling context,
+    // but it used to have one.
+    NS_ENSURE_TRUE(sgo || !hasHadScriptObject, PR_FALSE);
   }
 
   if (!document && !sgo) {
@@ -2194,6 +2361,8 @@ nsCxPusher::Push(nsISupports *aCurrentTarget)
 
     if (mScx) {
       cx = (JSContext *)mScx->GetNativeContext();
+      // Bad, no JSContext from script global object!
+      NS_ENSURE_TRUE(cx, PR_FALSE);
     }
   }
 
@@ -2223,6 +2392,7 @@ nsCxPusher::Push(nsISupports *aCurrentTarget)
 
     mScx = nsnull;
   }
+  return PR_TRUE;
 }
 
 void
@@ -2454,3 +2624,20 @@ nsContentUtils::RemoveJSGCRoot(void* aPtr)
 
   return NS_OK;
 }
+/* static */
+PRBool
+nsContentUtils::IsNativeAnonymous(nsIContent* aContent)
+{
+  while (aContent) {
+    nsIContent* bindingParent = aContent->GetBindingParent();
+    if (bindingParent == aContent) {
+      return PR_TRUE;
+    }
+    if (aContent->IsContentOfType(nsIContent::eTEXT)) {
+      aContent = aContent->GetParent();
+    } else {
+      aContent = bindingParent;
+    }
+  }
+  return PR_FALSE;
+}
